#pragma once

#include <coroutine>
#include <iostream>
#include <memory>
#include <optional>
#include <stdexcept>
#include <utility>
#include <vector>

#include <Game/Core.hpp>

namespace bt {
enum class task_result { failure, success };

class node_task {
public:
  class promise_type;

  using handle_type = std::coroutine_handle<promise_type>;

  class promise_type {
  public:
    std::optional<task_result> result;
    handle_type get_return_object() noexcept {
      return handle_type::from_promise(*this);
    }

    std::suspend_always initial_suspend() noexcept { return {}; }

    std::suspend_always final_suspend() noexcept { return {}; }

    void return_value(task_result r) noexcept { result = r; }

    void unhandled_exception() noexcept {
      std::terminate(); // TODO: handle exception
    }
  };

  class awaiter {
  public:
    awaiter(handle_type handle) noexcept : _handle(handle) {}

    bool await_ready() const noexcept { return false; }

    auto await_suspend(std::coroutine_handle<>) noexcept { return _handle; }

    task_result await_resume() noexcept { return *_handle.promise().result; }

  private:
    handle_type _handle;
  };

  node_task(handle_type handle) noexcept : _handle(handle) {}

  node_task(node_task &&other) noexcept : _handle(other._handle) {
    other._handle = nullptr;
  }

  ~node_task() {
    if (_handle) {
      _handle.destroy();
    }
  }

  node_task &operator=(node_task &&other) noexcept {
    if (&other != this) {
      if (_handle) {
        _handle.destroy();
      }
      _handle = other._handle;
      other._handle = nullptr;
    }
    return *this;
  }

  awaiter operator co_await() noexcept { return awaiter{_handle}; }

  void resume() noexcept { _handle.resume(); }

  bool done() const noexcept { return _handle.done(); }

  task_result result() noexcept { return *_handle.promise().result; }

  std::optional<task_result> try_result() noexcept {
    return _handle.promise().result;
  }

private:
  handle_type _handle;
};

using yield = std::suspend_always;

template <typename Context> class behavior_node {
public:
  virtual node_task tick(Context &ctx) noexcept = 0;

  virtual ~behavior_node() = default;
};

template <typename Context>
class composite_node : public behavior_node<Context> {
public:
  void add_child(std::unique_ptr<behavior_node<Context>> &&child) {
    _children.push_back(std::move(child));
  }

protected:
  std::vector<std::unique_ptr<behavior_node<Context>>> _children;
};

template <typename Context>
class sequence_node : public composite_node<Context> {
public:
  node_task tick(Context &ctx) noexcept override {
    std::size_t idx = 0;
    while (idx < this->_children.size()) {
      auto task = this->_children[idx]->tick(ctx);
      while (!task.done()) {
        task.resume();

        if (idx < this->_children.size() - 1)
          co_await yield{};
      }

      if (task.result() == task_result::failure) {
        co_return task_result::failure;
      }

      ++idx;
    }

    co_return task_result::success;
  }
};

template <typename Context>
class selector_node : public composite_node<Context> {
public:
  node_task tick(Context &ctx) noexcept override {
    for (std::size_t idx = 0; idx < this->_children.size(); ++idx) {
      auto task = this->_children[idx]->tick(ctx);
      while (!task.done()) {
        task.resume();
        co_await yield{};
      }

      if (task.result() == task_result::success) {
        co_return task_result::success;
      }
    }
    co_return task_result::failure;
  }
};

template <typename Context, typename TickFn>
class action_node : public behavior_node<Context> {
public:
  action_node(TickFn tick) noexcept : _tick(tick) {}

  node_task tick(Context &ctx) noexcept override { return _tick(ctx); }

private:
  TickFn _tick;
};

template <typename Context, typename ConditionFn>
class conditional_node : public behavior_node<Context> {
public:
  conditional_node(ConditionFn condition) noexcept : _condition(condition) {}

  node_task tick(Context &ctx) noexcept override {
    if (_condition(ctx)) {
      co_return task_result::success;
    }
    co_return task_result::failure;
  }

private:
  ConditionFn _condition;
};

template <typename Context>
class inverter_node : public behavior_node<Context> {
public:
  inverter_node(std::unique_ptr<behavior_node<Context>> child)
      : _child(std::move(child)) {}

  node_task tick(Context &ctx) noexcept override {
    auto task = _child->tick(ctx);
    while (!task.done()) {
      task.resume();
      co_await yield{};
    }

    if (task.result() == task_result::success) {
      co_return task_result::failure;
    }
    co_return task_result::success;
  }

private:
  std::unique_ptr<behavior_node<Context>> _child;
};

// Script
template <typename Context>
class terminate_script_node : public behavior_node<Context> {
public:
  node_task tick(Context &ctx) noexcept override {
    Terminate = true;
    co_return task_result::success;
  }
};

// Login
template<typename Context>
class login_player_node : public behavior_node<Context> {
public:
  node_task tick(Context &ctx) noexcept override {
    if (!Login::LoginPlayer())
      co_return task_result::failure;

    co_return task_result::success;
  }
};

// Mainscreen
template <typename Context>
class is_logged_in_node : public behavior_node<Context> {
public:
  constexpr is_logged_in_node() noexcept = default;

  node_task tick(Context &ctx) noexcept override {
    if (!Mainscreen::IsLoggedIn())
      co_return task_result::failure;
    
    co_return task_result::success;
  }
};

// Player
template <typename Context>
class is_animating_node : public behavior_node<Context> {
public:
  std::int32_t anim_id;

  is_animating_node(std::int32_t anim_id) : anim_id(anim_id) {}

  node_task tick(Context &ctx) noexcept override {
    const auto player = Players::GetLocal();
    if (!player)
      co_return task_result::failure;

    if (player.GetAnimationID() != anim_id)
      co_return task_result::failure;

    co_return task_result::success;
  }
};

// Inventory
template <typename Context, typename ItemID>
  requires std::same_as<std::int32_t, ItemID> ||
           std::same_as<std::string, ItemID> ||
           std::same_as<std::regex, ItemID> ||
           std::same_as<std::vector<std::int32_t>, ItemID> ||
           std::same_as<std::vector<std::string>, ItemID>
class inventory_contains_node : public behavior_node<Context> {
public:
  ItemID item_id;

  inventory_contains_node(ItemID item_id) : item_id(item_id) {}

  node_task tick(Context &ctx) noexcept override {
    if (!Inventory::Contains(item_id))
      co_return task_result::failure;

    co_return task_result::success;
  }
};

// Magic
template <typename Context>
class select_spell_node : public behavior_node<Context> {
public:
  Magic::SPELL spell;

  select_spell_node(Magic::SPELL spell) : spell(spell) {}

  node_task tick(Context &ctx) noexcept override {
    if (!Magic::SelectSpell(spell))
      co_return task_result::failure;

    co_return task_result::success;
  }
};

template <typename Context>
class spell_selected_node : public behavior_node<Context> {
public:
  Magic::SPELL spell;

  spell_selected_node(Magic::SPELL spell) : spell(spell) {}

  node_task tick(Context &ctx) noexcept override {
    if (!Magic::IsSpellSelected(spell))
      co_return task_result::failure;

    co_return task_result::success;
  }
};

template <typename Context>
class cast_spell_node : public behavior_node<Context> {
public:
  Magic::SPELL spell;

  cast_spell_node(Magic::SPELL spell) : spell(spell) {}

  node_task tick(Context &ctx) noexcept override {
    if (!Magic::CastSpell(spell))
      co_return task_result::failure;

    co_return task_result::success;
  }
};

// Interact
template <typename Context, typename Interactable> class interact_node;

template <typename Context>
class interact_node<Context, Interactable::Item>
    : public behavior_node<Context> {
public:
  std::string name;
  std::string action;

  interact_node(const std::string &name, const std::string &action)
      : name(name), action(action) {}

  node_task tick(Context &ctx) noexcept override {
    const auto item = Inventory::GetItem(name);
    if (!item)
      co_return task_result::failure;

    if (!item.Interact(action))
      co_return task_result::failure;

    co_return task_result::success;
  }
};

template <typename Context> class tree_builder {
public:
  tree_builder() noexcept : _root(nullptr) {}

  tree_builder &begin_sequence() {
    _node_stack.push_back(std::make_unique<sequence_node<Context>>());
    _children.emplace_back();
    return *this;
  }

  tree_builder &end_sequence() {
    if (_node_stack.empty()) {
      throw std::runtime_error("Mismatched begin_sequence/end_sequence");
    }

    auto node = std::move(_node_stack.back());
    _node_stack.pop_back();
    auto children = std::move(_children.back());
    _children.pop_back();

    for (auto &&child : children) {
      static_cast<composite_node<Context> *>(node.get())
          ->add_child(std::move(child));
    }

    if (_node_stack.empty()) {
      _root = std::move(node);
    } else {
      _children.back().push_back(std::move(node));
    }

    return *this;
  }

  tree_builder &begin_selector() {
    _node_stack.push_back(std::make_unique<selector_node<Context>>());
    _children.emplace_back();
    return *this;
  }

  tree_builder &end_selector() {
    if (_node_stack.empty()) {
      throw std::runtime_error("Mismatched begin_selector/end_selector");
    }

    auto node = std::move(_node_stack.back());
    _node_stack.pop_back();
    auto children = std::move(_children.back());
    _children.pop_back();

    for (auto &&child : children) {
      static_cast<composite_node<Context> *>(node.get())
          ->add_child(std::move(child));
    }

    if (_node_stack.empty()) {
      _root = std::move(node);
    } else {
      _children.back().push_back(std::move(node));
    }

    return *this;
  }

  template <typename TickFn> tree_builder &action(TickFn tick) {
    if (_node_stack.empty()) {
      throw std::runtime_error("Action must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<action_node<Context, TickFn>>(tick));
    return *this;
  }

  template <typename ConditionFn>
  tree_builder &conditional(ConditionFn condition) {
    if (_node_stack.empty()) {
      throw std::runtime_error(
          "Condition must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<conditional_node<Context, ConditionFn>>(condition));
    return *this;
  }

  tree_builder &invert() {
    if (_node_stack.empty()) {
      throw std::runtime_error("Invert must be within a sequence or selector");
    }

    if (_children.back().empty()) {
      throw std::runtime_error("Invert requires a preceding node");
    }

    auto child = std::move(_children.back().back());
    _children.back().pop_back();
    _children.back().push_back(
        std::make_unique<inverter_node<Context>>(std::move(child)));

    return *this;
  }

  tree_builder& terminate_script() {
    if (_node_stack.empty()) {
      throw std::runtime_error("Terminate must be within a sequence or selector");
    }

    _children.back().push_back(std::make_unique<terminate_script_node<Context>>());
    return *this;
  }

  tree_builder &login_player() {
    if (_node_stack.empty()) {
      throw std::runtime_error("Leaf must be within a sequence or selector");
    }

    _children.back().push_back(std::make_unique<login_player_node<Context>>());
    return *this;
  }

  tree_builder &is_animating(std::int32_t anim_id) {
    if (_node_stack.empty()) {
      throw std::runtime_error("Leaf must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<is_animating_node<Context>>(anim_id));
    return *this;
  }

  tree_builder &is_logged_in() {
    if (_node_stack.empty()) {
      throw std::runtime_error("Leaf must be within a sequence or selector");
    }

    _children.back().push_back(std::make_unique<is_logged_in_node<Context>>());
    return *this;
  }

  template <typename ItemID>
    requires std::same_as<std::int32_t, ItemID> ||
             std::same_as<std::string, ItemID> ||
             std::same_as<std::regex, ItemID> ||
             std::same_as<std::vector<std::int32_t>, ItemID> ||
             std::same_as<std::vector<std::string>, ItemID>
  tree_builder &inventory_contains(ItemID item_id) {
    if (_node_stack.empty()) {
      throw std::runtime_error("Leaf must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<inventory_contains_node<Context, ItemID>>(item_id));
    return *this;
  }

  tree_builder &select_spell(Magic::SPELL spell) {
    if (_node_stack.empty()) {
      throw std::runtime_error("Leaf must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<select_spell_node<Context>>(spell));
    return *this;
  }

  tree_builder &spell_selected(Magic::SPELL spell) {
    if (_node_stack.empty()) {
      throw std::runtime_error("Leaf must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<spell_selected_node<Context>>(spell));
    return *this;
  }

  tree_builder &cast_spell(Magic::SPELL spell) {
    if (_node_stack.empty()) {
      throw std::runtime_error("Leaf must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<cast_spell_node<Context>>(spell));
    return *this;
  }

  template <typename Interactable>
  tree_builder &interact(const std::string &name, const std::string &action) {
    if (_node_stack.empty()) {
      throw std::runtime_error("Leaf must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<interact_node<Context, Interactable>>(name, action));
    return *this;
  }

  std::unique_ptr<behavior_node<Context>> build() {
    if (!_node_stack.empty()) {
      throw std::runtime_error("Mismatched begin/end");
    }

    return std::move(_root);
  }

private:
  std::unique_ptr<behavior_node<Context>> _root;
  std::vector<std::unique_ptr<behavior_node<Context>>> _node_stack;
  std::vector<std::vector<std::unique_ptr<behavior_node<Context>>>> _children;
};

} // namespace bt